import pathlib

import pydantic


@pydantic.validate_call
def create_file_tree(directory: pathlib.Path, structure: dict[str, dict | str]) -> None:
    """
    Creates a directory and file tree based on a dictionary structure.

    Each key in the dictionary represents a file or directory name.
    If the value is a dictionary, it represents a subdirectory with its own structure.
    If the value is a string, the key is taken to represent a file with this value as its content.

    Parameters
    ----------
    directory : pathlib.Path
        The root directory where the tree will be created.
    structure : dict
        A nested dictionary representing the directory and file structure.
    """
    for key, value in structure.items():
        subpath = directory / key
        if isinstance(value, dict):
            subpath.mkdir(exist_ok=True)
            create_file_tree(directory=subpath, structure=value)
        elif isinstance(value, str):
            subpath.write_text(data=value)



import pydantic

from .._converters._dataset_converter import DatasetConverter
from .._converters._run_config import RunConfig


@pydantic.validate_call
def convert_nwb_dataset(
    *,
    nwb_paths: list[pydantic.FilePath | pydantic.DirectoryPath] = pydantic.Field(min_length=1),
    run_config: RunConfig = pydantic.Field(default_factory=RunConfig),
) -> DatasetConverter:
    """
    Convert a dataset of NWB files to a BIDS dataset.

    Parameters
    ----------
    nwb_paths : any iterable of file or directory paths
        An iterable of NWB file paths and directories containing NWB files.
    run_config : RunConfig, optional
        The configuration for this conversion run.

    Returns
    -------
    converter : DatasetConverter
        The DatasetConverter used to perform the conversion.
        Contains notifications and other contextual information about the conversion process.
    """
    converter = DatasetConverter.from_nwb_paths(nwb_paths=nwb_paths, run_config=run_config)
    converter.extract_metadata()
    converter.convert_to_bids_dataset()

    return converter



"""Configuration file for the doctests."""
import json
import pathlib
import typing

import pytest

import nwb2bids


# Doctest directories
@pytest.fixture(autouse=True)
def add_data_space(doctest_namespace: dict[str, typing.Any], tmp_path: pathlib.Path):
    doctest_namespace["path_to_some_directory"] = pathlib.Path(tmp_path)

    nwb2bids.testing.generate_ephys_tutorial(mode="file")
    nwb2bids.testing.generate_ephys_tutorial(mode="dataset")

    tutorial_directory = nwb2bids.testing.get_tutorial_directory() / "ephys_tutorial_file"
    additional_metadata_file_path = tutorial_directory / "metadata.json"

    additional_metadata = {
        "dataset_description": {
            "Name": "My Custom BIDS Dataset",
            "BIDSVersion": "1.8.0",
            "Authors": ["First Last", "Second Author"]
        }
    }
    additional_metadata_file_path.write_text(data=json.dumps(obj=additional_metadata))



import datetime
import json
import os
import pathlib
import shutil
import sys
import uuid

import py.path
import pynwb
import pynwb.file
import pynwb.testing.mock.ecephys
import pynwb.testing.mock.file
import pytest

import nwb2bids

# These tests fail on Windows GitHub CI due to git-annex adjusted branch issues
# See https://github.com/con/nwb2bids/pull/213 for failure output
pytest_mark_xfail_windows_github_ci = pytest.mark.xfail(
    sys.platform == "win32" and os.environ.get("GITHUB_ACTIONS", "").lower() == "true",
    reason="git-annex adjusted branch fails on Windows GitHub CI runners",
    strict=False,
)

# TODO: add ndx-events
# TODO: add DynamicTable's in acquisition with *_time columns


def _make_minimal_nwbfile(session_id: str = "456", subject_species: str = "Mus musculus") -> pynwb.NWBFile:
    """Creates a minimal NWB file for testing purposes."""
    nwbfile = pynwb.testing.mock.file.mock_NWBFile(session_id=session_id)

    subject = pynwb.file.Subject(
        subject_id="123",
        species=subject_species,
        sex="M",
    )
    nwbfile.subject = subject

    return nwbfile


@pytest.fixture()
def temporary_bids_directory(tmpdir: py.path.local) -> pathlib.Path:
    """Creates a temporary BIDS directory for testing purposes."""
    return pathlib.Path(tmpdir)


@pytest.fixture(scope="session")
def testing_files_directory(tmp_path_factory: pytest.TempPathFactory) -> pathlib.Path:
    """
    Creates a common temporary directory for all testing files.

    Testing files may be nested into subdirectories to test localization logic.
    """
    testing_files_directory = tmp_path_factory.mktemp(basename="nwb2bids_testing_files", numbered=False)
    return testing_files_directory


@pytest.fixture(scope="session")
def additional_metadata_file_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    additional_metadata_dictionary = {
        "dataset_description": {
            "Name": "test",
            "Description": "TODO",
            "BIDSVersion": "1.10",
            "DatasetType": "raw",
            "License": "CC-BY-4.0",
            "Authors": ["Cody Baker", "Yaroslav Halchenko"],
        }
    }

    additional_metadata_subdirectory = testing_files_directory / "additional_metadata"
    additional_metadata_subdirectory.mkdir(exist_ok=True)
    file_path = additional_metadata_subdirectory / "test_additional_metadata.json"
    with file_path.open(mode="w") as file_stream:
        json.dump(obj=additional_metadata_dictionary, fp=file_stream)

    return file_path


@pytest.fixture(scope="session")
def additional_metadata_with_generated_by_file_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    additional_metadata_dictionary = {
        "dataset_description": {
            "Name": "test",
            "Description": "Dataset with user-provided GeneratedBy",
            "BIDSVersion": "1.10",
            "DatasetType": "raw",
            "License": "CC-BY-4.0",
            "Authors": ["Cody Baker", "Yaroslav Halchenko"],
            "GeneratedBy": [
                {
                    "Name": "custom-pipeline",
                    "Version": "1.0.0",
                    "Description": "Custom data processing pipeline",
                    "CodeURL": "https://github.com/example/custom-pipeline",
                }
            ],
        }
    }

    additional_metadata_subdirectory = testing_files_directory / "additional_metadata"
    additional_metadata_subdirectory.mkdir(exist_ok=True)
    file_path = additional_metadata_subdirectory / "test_additional_metadata_with_generated_by.json"
    with file_path.open(mode="w") as file_stream:
        json.dump(obj=additional_metadata_dictionary, fp=file_stream)

    return file_path


@pytest.fixture(scope="session")
def minimal_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    A minimally valid NWB file for testing purposes.

    Does not contain any additional metadata beyond what is required for initialization of converter and metadata
    classes.
    """
    nwbfile = _make_minimal_nwbfile()

    minimal_subdirectory = testing_files_directory / "minimal"
    minimal_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = minimal_subdirectory / "minimal.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def minimal_mismatch_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    Prior to PR #???, it was possible to create `participants.tsv` files with duplicated `participant_id` values.

    This was because of the way dataframes were merged during the creation of the participants table.
    """
    nwbfile = _make_minimal_nwbfile(session_id="4567", subject_species="mouse")

    minimal_subdirectory = testing_files_directory / "minimal_mismatch"
    minimal_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = minimal_subdirectory / "minimal_mismatch.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def ecephys_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = _make_minimal_nwbfile(session_id="789")

    pynwb.testing.mock.ecephys.mock_ElectricalSeries(name="test_electrical_series", nwbfile=nwbfile)

    ecephys_subdirectory = testing_files_directory / "ecephys"
    ecephys_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = ecephys_subdirectory / "ecephys.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def trials_events_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = _make_minimal_nwbfile()

    trials = nwb2bids.testing.mock_trials_table(nwbfile=nwbfile)
    nwbfile.trials = trials

    events_subdirectory = testing_files_directory / "trials"
    events_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = events_subdirectory / "trials.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def epochs_events_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = _make_minimal_nwbfile()

    epochs = nwb2bids.testing.mock_epochs_table()
    nwbfile.epochs = epochs

    events_subdirectory = testing_files_directory / "epochs"
    events_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = events_subdirectory / "epochs.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def multiple_events_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = _make_minimal_nwbfile()

    trials = nwb2bids.testing.mock_trials_table()
    nwbfile.trials = trials

    epochs = nwb2bids.testing.mock_epochs_table()
    nwbfile.epochs = epochs

    time_intervals = nwb2bids.testing.mock_time_intervals()
    nwbfile.add_acquisition(time_intervals)

    events_subdirectory = testing_files_directory / "multiple_events"
    events_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = events_subdirectory / "multiple_events.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def directory_with_multiple_nwbfiles(testing_files_directory: pathlib.Path) -> pathlib.Path:
    multiple_nwbfiles_subdirectory = testing_files_directory / "multiple_nwbfiles"
    multiple_nwbfiles_subdirectory.mkdir(exist_ok=True)

    for session_index in range(2):
        nwbfile = _make_minimal_nwbfile(session_id=f"session-{session_index}")

        nwbfile_path = multiple_nwbfiles_subdirectory / f"session_{session_index}.nwb"
        with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
            file_stream.write(nwbfile)

    return multiple_nwbfiles_subdirectory


@pytest.fixture(scope="session")
def mock_datalad_dataset(testing_files_directory: pathlib.Path, minimal_nwbfile_path: pathlib.Path) -> pathlib.Path:
    """
    A mock datalad dataset for testing purposes.

    Original globbing pattern was too broad and found files within the git-annex.
    """
    dataset_subdirectory = testing_files_directory / "mock_datalad_dataset"
    dataset_subdirectory.mkdir(exist_ok=True)

    annex_filename = "MD5E-s14336--bd0eed310fabd903a2635186e06b6a43.nwb"
    structure = {
        ".datalad": {
            ".gitattributes": "config annex.largefiles=nothing\n",
            "config": '[datalad "dataset"]\n\tid = NOT-A-REAL-DATALAD-DATASET',
        },
        ".git": {"annex": {"objects": {"abc": {"def": {annex_filename: {annex_filename: ""}}}}}},
    }
    nwb2bids.testing.create_file_tree(directory=dataset_subdirectory, structure=structure)

    content_file_path = dataset_subdirectory / ".git/annex/objects/abc/def" / annex_filename / annex_filename
    shutil.copy2(src=minimal_nwbfile_path, dst=content_file_path)

    annexed_file_path = dataset_subdirectory / "minimal.nwb"
    annexed_file_path.symlink_to(target=content_file_path)

    return dataset_subdirectory


# Problematic test cases
@pytest.fixture(scope="session")
def problematic_nwbfile_path_1(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    An NWB file with 'problematic' metadata for testing the messaging feature.
    """
    nwbfile = pynwb.NWBFile(
        identifier="not a UUID",
        session_id="problematic1",
        session_description="",
        session_start_time=datetime.datetime.now().astimezone(),
    )

    subject = pynwb.file.Subject(
        subject_id="bad_subject_id",
        species="not a latin species, nor a ncbi taxon link or id",
        sex="bad sex specifier",
    )
    nwbfile.subject = subject

    problematic_subdirectory = testing_files_directory / "problematic"
    problematic_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = problematic_subdirectory / "problematic1.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def problematic_nwbfile_path_2(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    A second NWB file with 'problematic' metadata for testing the messaging feature.
    """
    nwbfile = pynwb.NWBFile(
        identifier="not a UUID",
        session_id="#problematic!2~",
        session_description="",
        session_start_time=datetime.datetime.now().astimezone(),
    )

    subject = pynwb.file.Subject(
        subject_id="bad subject id",
    )
    nwbfile.subject = subject

    problematic_subdirectory = testing_files_directory / "problematic"
    problematic_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = problematic_subdirectory / "problematic2.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def problematic_nwbfile_path_3(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    A third NWB file with 'problematic' metadata (entirely missing entities) for testing the messaging feature.
    """
    nwbfile = pynwb.NWBFile(
        identifier="not a UUID",
        session_id="problematic3",
        session_description="",
        session_start_time=datetime.datetime.now().astimezone(),
    )

    problematic_subdirectory = testing_files_directory / "problematic"
    problematic_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = problematic_subdirectory / "problematic3.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def problematic_nwbfile_path_4(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    A fourth NWB file with less problematic metadata corresponding only to low-level 'info' events.
    """
    nwbfile = pynwb.NWBFile(
        identifier=str(uuid.uuid4()),
        session_id="problematic4",
        session_description="",
        session_start_time=datetime.datetime.now().astimezone(),
    )
    subject = pynwb.file.Subject(
        subject_id="123",
        species="Mus musculus",
        sex="M",
    )
    nwbfile.subject = subject

    device = pynwb.testing.mock.ecephys.mock_Device(name="DeviceWithoutDescription", description=None, nwbfile=nwbfile)
    group = pynwb.testing.mock.ecephys.mock_ElectrodeGroup(device=device, nwbfile=nwbfile)
    nwbfile.add_electrode(group=group, location="unknown")

    problematic_subdirectory = testing_files_directory / "problematic"
    problematic_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = problematic_subdirectory / "problematic4.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def problematic_nwbfile_path_missing_session_id(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = pynwb.testing.mock.file.mock_NWBFile(session_id=None)

    subject = pynwb.file.Subject(
        subject_id="123",
        species="Mus musculus",
        sex="M",
    )
    nwbfile.subject = subject

    events_subdirectory = testing_files_directory / "missing_session_id"
    events_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = events_subdirectory / "missing_session_id.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path



import importlib.metadata
import json
import typing

import pydantic
import typing_extensions

from ._base_metadata_model import BaseMetadataModel


class GeneratedByItem(BaseMetadataModel):
    """
    Schema for a single GeneratedBy entry in BIDS dataset_description.json.

    Represents provenance information about a pipeline or process that generated the dataset.

    See Also
    --------
    https://bids-specification.readthedocs.io/en/stable/glossary.html#generatedby-metadata
    """

    Name: str = pydantic.Field(
        description="Name of the pipeline or process that generated the outputs.",
    )
    Version: str = pydantic.Field(
        description="Version of the pipeline.",
    )
    Description: str = pydantic.Field(
        description="Plain-text description of the pipeline or process that generated the outputs.",
    )
    CodeURL: str = pydantic.Field(
        description="URL where the code used to generate the dataset may be found.",
    )


class GeneratedByNwb2bids(GeneratedByItem):
    """
    nwb2bids-specific GeneratedBy entry with defaults for the nwb2bids pipeline.
    """

    Name: str = pydantic.Field(default="nwb2bids")
    Version: str = pydantic.Field(default_factory=lambda: importlib.metadata.version(distribution_name="nwb2bids"))
    Description: str = pydantic.Field(default="Tool to reorganize NWB files into a BIDS directory layout.")
    CodeURL: str = pydantic.Field(default="https://github.com/con/nwb2bids")


class DatasetDescription(BaseMetadataModel):
    """
    Schema for the dataset description in BIDS format.
    """

    Name: str = pydantic.Field(description="Name of the dataset.")
    BIDSVersion: str | None = pydantic.Field(
        description="The version of the BIDS standard that was used.",
        pattern=r"^\d+\.\d+(?:\.\d+)?$",
    )
    Description: str | None = pydantic.Field(
        description="Description of the dataset.",
        default=None,
    )
    DatasetType: typing.Literal["raw", "derivative"] = pydantic.Field(
        description="The interpretation of the dataset.",
        default="raw",
    )
    Authors: list[str] | None = pydantic.Field(
        description="List of individuals who contributed to the creation/curation of the dataset.",
        default=None,
    )
    License: typing.Literal["CC-BY-4.0", "CC0-1.0"] | None = pydantic.Field(
        description="License under which the dataset is released.",
        default=None,
    )
    GeneratedBy: list[GeneratedByItem] | None = pydantic.Field(
        description="Provenance information - pipelines that generated this dataset.",
        default=None,
    )

    def model_post_init(self, context: typing.Any, /) -> None:
        generated_by_nwb2bids = GeneratedByNwb2bids()
        if self.GeneratedBy is None:
            self.GeneratedBy = [generated_by_nwb2bids]
        else:
            self.GeneratedBy.append(generated_by_nwb2bids)

    @pydantic.model_validator(mode="after")
    def validate_exactly_one_nwb2bids(self) -> typing_extensions.Self:
        if self.GeneratedBy is not None:
            nwb2bids_count = sum(1 for item in self.GeneratedBy if item.Name == "nwb2bids")
            if nwb2bids_count != 1:
                raise ValueError(f"GeneratedBy must contain exactly one nwb2bids entry, found {nwb2bids_count}")
        return self

    @classmethod
    @pydantic.validate_call
    def from_file_path(cls, file_path: pydantic.FilePath) -> typing_extensions.Self | None:
        """
        Load the BIDS dataset description from a JSON file.

        Parameters
        ----------
        file_path : FilePath
            Path to the JSON file containing the dataset metadata.

        Returns
        -------
        BidsDatasetMetadata or None
            An instance of DatasetDescription populated with data from the file.
        """
        with file_path.open(mode="r") as file_stream:
            dictionary = json.load(fp=file_stream)

        dataset_description_content = dictionary.get("dataset_description", None)
        if dataset_description_content is None:
            return None

        dataset_metadata = cls(**dataset_description_content)
        return dataset_metadata



import typing

import pydantic

from .._inspection._inspection_result import Category, InspectionResult, Severity
from ..bids_models import DatasetDescription
from ..bids_models._model_globals import _BIDS_RRID


def get_bids_dataset_description(dandiset) -> tuple[DatasetDescription | None, list[InspectionResult]]:
    valid_or_raw: typing.Literal["raw", "valid"] = "valid"
    try:
        metadata = dandiset.get_metadata()
    except pydantic.ValidationError:
        raw_metadata = dandiset.get_raw_metadata()
        valid_or_raw = "raw"

    messages = []
    if valid_or_raw == "valid":
        dataset_description, internal_messages = _get_dataset_description_from_valid_dandiset_metadata(
            metadata=metadata
        )
        messages.extend(internal_messages)
    else:
        dataset_description, internal_messages = _get_dataset_description_from_invalid_dandiset_metadata(
            raw_metadata=raw_metadata
        )
        messages.extend(internal_messages)

        message = InspectionResult(
            title="INFO: invalid Dandiset metadata",
            reason="This Dandiset has invalid metadata.",
            solution="Required dataset description fields are inferred from the raw metadata instead.",
            category=Category.INTERNAL_ERROR,
            severity=Severity.INFO,
        )
        messages.append(message)

    return dataset_description, messages


def _get_dataset_description_from_valid_dandiset_metadata(
    metadata: typing.Any,
) -> tuple[DatasetDescription | None, list[InspectionResult]]:
    dataset_description_kwargs = dict()

    dataset_description_kwargs["Name"] = metadata.name
    dataset_description_kwargs["BIDSVersion"] = "1.10"

    messages = []
    if any(data_standard.identifier == _BIDS_RRID for data_standard in metadata.assetsSummary.dataStandard):
        reason = (
            "Dandiset is already organized to the BIDS standard. If only a partial conversion is desired, "
            "please raise an issue on https://github.com/con/nwb2bids/issues/new to discuss the use case."
        )
        message = InspectionResult(
            title="Dandiset is already BIDS",
            reason=reason,
            solution="Skip the conversion of this Dandiset.",
            category=Category.INTERNAL_ERROR,
            severity=Severity.ERROR,
        )
        messages.append(message)
        return None, messages

    if metadata.description is not None:
        dataset_description_kwargs["Description"] = metadata.description
    if metadata.contributor is not None:
        dataset_description_kwargs["Authors"] = [
            contributor.name
            for contributor in metadata.contributor
            for role in contributor.roleName
            if role.value == "dcite:Author"
        ]
    if metadata.license is not None:
        dataset_description_kwargs["License"] = metadata.license[0].value.split(":")[1]

    dataset_description = DatasetDescription(**dataset_description_kwargs)
    return dataset_description, messages


def _get_dataset_description_from_invalid_dandiset_metadata(
    raw_metadata: dict[str, typing.Any],
) -> tuple[DatasetDescription | None, list[InspectionResult]]:
    dataset_description_kwargs = dict()

    dandiset_identifier = raw_metadata.get("identifier", "??????")
    dataset_description_kwargs["Name"] = raw_metadata.get("name", f"DANDI Archive Dandiset {dandiset_identifier}")
    dataset_description_kwargs["BIDSVersion"] = "1.10"

    assets_summary = raw_metadata.get("assetsSummary", dict())
    data_standards = assets_summary.get("dataStandard", [])

    messages = []
    if any(data_standard.get("identifier", "") == _BIDS_RRID for data_standard in data_standards):
        reason = (
            "Dandiset is already organized to the BIDS standard. If only a partial conversion is desired, "
            "please raise an issue on https://github.com/con/nwb2bids/issues/new to discuss the use case."
        )
        message = InspectionResult(
            title="Dandiset is already BIDS",
            reason=reason,
            solution="Skip the conversion of this Dandiset.",
            category=Category.INTERNAL_ERROR,
            severity=Severity.ERROR,
        )
        messages.append(message)
        return None, messages

    dandiset_description = raw_metadata.get("description", None)
    if dandiset_description is not None:
        dataset_description_kwargs["Description"] = dandiset_description

    bids_authors = []
    dandiset_contributors = raw_metadata.get("contributor", [])
    for contributor in dandiset_contributors:
        contributor_name = contributor.get("name", None)
        is_author = "dcite:Author" in contributor.get("roleName", [])
        if contributor_name is not None and is_author:
            bids_authors.append(contributor_name)

    if any(bids_authors):
        dataset_description_kwargs["Authors"] = bids_authors

    license = raw_metadata.get("license", [])
    if len(license) > 0:
        if len(license) == 1:
            dataset_description_kwargs["License"] = license[0].split(":")[-1]
        else:
            message = InspectionResult(
                title="WARNING: multiple licenses not supported",
                reason="DANDI metadata supports multiple licenses, but BIDS only supports one license.",
                solution="Manually specify the license in the dataset_description.json file after conversion.",
                category=Category.SCHEMA_INVALIDATION,
                severity=Severity.WARNING,
            )
            messages.append(message)

    dataset_description = DatasetDescription(**dataset_description_kwargs)
    return dataset_description, messages



"""Integration tests for the primary `nwb2bids convert` CLI."""

import pathlib
import subprocess


def test_problematic_cli_error_messages(
    problematic_nwbfile_path_1: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    command = f"nwb2bids convert {problematic_nwbfile_path_1} -o {temporary_bids_directory}"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert result.returncode == 0

    expected_message = [
        "",
        "BIDS dataset was not successfully created!",
        "Some errors were encountered during conversion.",
        "The first 3 of 4 are shown below:",
        "",
        "",
        "\t- Participant species is not a proper Latin binomial or NCBI Taxonomy id.",
        "\t- The participant ID contains invalid characters. BIDS allows only the "
        "plus sign to be used as a separator in the subject entity label. "
        "Underscores, dashes, spaces, slashes, and other special characters "
        "(including #) are expressly forbidden.",
        "\t- Participant sex is not one of the allowed patterns by BIDS.",
        "",
    ]
    actual_lines = result.stdout.decode(encoding="utf-8").splitlines()
    assert expected_message == actual_lines[:-2]
    assert "Please review the full notifications report at" in actual_lines[-2]
    assert result.stderr == b""


def test_problematic_cli_critical_messages(
    problematic_nwbfile_path_3: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    command = f"nwb2bids convert {problematic_nwbfile_path_3} -o {temporary_bids_directory}"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert result.returncode == 0

    expected_message = ["", "BIDS dataset was successfully created, but may not be valid!", ""]
    actual_lines = result.stdout.decode(encoding="utf-8").splitlines()
    assert expected_message == actual_lines[:-2]
    assert "Please review the full notifications report at" in actual_lines[-2]
    assert result.stderr == b""


def test_problematic_cli_info_messages(
    problematic_nwbfile_path_4: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    command = f"nwb2bids convert {problematic_nwbfile_path_4} -o {temporary_bids_directory}"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert result.returncode == 0

    expected_message = [
        "",
        "BIDS dataset was successfully created!",
        "1 suggestion for improvement was found during conversion.",
        "",
    ]
    actual_lines = result.stdout.decode(encoding="utf-8").splitlines()
    assert expected_message == actual_lines[:-2]
    assert "Please review the full notifications report at" in actual_lines[-2]
    assert result.stderr == b""


def test_cli_success(minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    command = f"nwb2bids convert {minimal_nwbfile_path} -o {temporary_bids_directory}"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert result.returncode == 0

    expected_message = ["", "BIDS dataset was successfully created!", ""]
    actual_lines = result.stdout.decode(encoding="utf-8").splitlines()
    assert expected_message == actual_lines
    assert result.stderr == b""


def test_problematic_cli_silent(problematic_nwbfile_path_1: pathlib.Path, temporary_bids_directory: pathlib.Path):
    command = f"nwb2bids convert {problematic_nwbfile_path_1} -o {temporary_bids_directory} --silent"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert result.returncode == 0
    assert result.stdout == b""
    assert result.stderr == b""



import datetime
import pathlib
import typing

import pydantic

from .._core._file_mode import _determine_file_mode
from .._core._home import _get_nwb2bids_home_directory
from .._core._validate_existing_bids import _validate_bids_directory


def _generate_run_id() -> str:
    """
    Generate a unique run ID based on the current date and time.

    Returns
    -------
    run_id : str
        On each unique run of `nwb2bids`, a run ID is generated.
        Set this option to override this to any identifying string.
        This ID is used in the naming of the notification and sanitization reports saved to your cache directory.
        The default ID uses runtime timestamp information of the form "date-%Y%m%d_time-%H%M%S."
    """
    run_id = datetime.datetime.now().strftime("datetime-%Y%m%d%H%M%S")
    return run_id


class RunConfig(pydantic.BaseModel):
    """
    Specifies configuration options for a single run of NWB to BIDS conversion.

    bids_directory : directory path
        The path to the directory where the BIDS dataset will be created.
        Defaults to the current working directory and checks if it is either empty or a BIDS dataset.
    additional_metadata_file_path : file path, optional
        The path to a YAML file containing additional metadata not included within the NWB files
        that you wish to include in the BIDS dataset.
    file_mode : one of "move", "copy", or "symlink"
        Specifies how to handle the NWB files when converting to BIDS format.
            - "move": Move the files to the BIDS directory.
            - "copy": Copy the files to the BIDS directory.
            - "symlink": Create symbolic links to the files in the BIDS directory.
            - if not specified, decide between all the above based on the system,
              with preference for linking when possible.
    cache_directory : directory path
        The directory where run specific files (e.g., notifications, sanitization reports) will be stored.
        Defaults to `~/.nwb2bids`.
    run_id : str
        On each unique run of nwb2bids, a run ID is generated.
        Set this option to override this to any identifying string.
        This ID is used in the naming of the files saved to your run directory.
        The default ID uses runtime timestamp information of the form "date-%Y%m%d_time-%H%M%S."
    """

    bids_directory: typing.Annotated[
        pathlib.Path,
        pydantic.Field(default_factory=pathlib.Path.cwd),
        pydantic.AfterValidator(_validate_bids_directory),
    ]
    additional_metadata_file_path: pydantic.FilePath | None = None
    file_mode: typing.Annotated[
        typing.Literal["move", "copy", "symlink"], pydantic.Field(default_factory=_determine_file_mode)
    ]
    cache_directory: typing.Annotated[
        pydantic.DirectoryPath, pydantic.Field(default_factory=_get_nwb2bids_home_directory)
    ]
    run_id: typing.Annotated[str, pydantic.Field(default_factory=_generate_run_id)]
    _nwb2bids_directory: pathlib.Path = pydantic.PrivateAttr()

    model_config = pydantic.ConfigDict(
        frozen=True,  # Make the model immutable
        validate_default=True,  # Validate default values as well
    )

    def model_post_init(self, context: typing.Any, /) -> None:
        self._nwb2bids_directory = self.bids_directory / ".nwb2bids"

    @pydantic.computed_field
    @property
    def notifications_file_path(self) -> pathlib.Path:
        """The file path leading to a human-readable notifications report."""
        notifications_file_path = self._nwb2bids_directory / f"{self.run_id}_notifications.txt"
        return notifications_file_path

    @pydantic.computed_field
    @property
    def notifications_json_file_path(self) -> pathlib.Path:
        """The file path leading to a JSON dump of the notifications."""
        notifications_file_path = self._nwb2bids_directory / f"{self.run_id}_notifications.json"
        return notifications_file_path



import functools
import pathlib
import warnings

import pynwb

ignored_pynwb_message_patterns = [
    "No cached namespaces found in .+",
    "Ignoring cached namespace .+",
    "Ignoring the following cached .+",
    "Stimulus description 'NA' for .+",
    "Use of icephys_filtering is deprecated and will be removed .+",
    "Date is missing timezone .+",
]
for message in ignored_pynwb_message_patterns:
    warnings.filterwarnings(action="ignore", message=message)


@functools.cache
def cache_read_nwb(file_path: pathlib.Path) -> pynwb.NWBFile:
    """Cache the read operation per NWB file path to speed up repeated calls."""
    # Always resolve in case of symlinks, mostly relevant to Windows (i.e., from DataLad most likely)
    resolved_path = file_path.resolve()
    nwbfile = pynwb.read_nwb(path=resolved_path)
    return nwbfile



"""Integration tests for the primary `convert_nwb_dataset` function with events data."""

import json
import pathlib

import pandas

import nwb2bids


def test_trials_events(trials_events_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    nwb_paths = [trials_events_nwbfile_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
                "sub-123_ses-456_events.tsv",
                "sub-123_ses-456_events.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    tsv_file_path = temporary_bids_directory / "sub-123" / "ses-456" / "ecephys" / "sub-123_ses-456_events.tsv"
    actual_dataframe = pandas.read_csv(filepath_or_buffer=tsv_file_path, sep="\t")
    expected_dataframe = pandas.DataFrame(
        {
            "onset": {0: 0.0, 1: 2.0, 2: 5.0, 3: 5.5},
            "duration": {0: 1.0, 1: 1.0, 2: 0.5, 3: 0.5},
            "nwb_table": {0: "trials", 1: "trials", 2: "trials", 3: "trials"},
            "trial_condition": {0: "A", 1: "B", 2: "C", 3: "D"},
        }
    )
    pandas.testing.assert_frame_equal(left=actual_dataframe, right=expected_dataframe)

    json_file_path = temporary_bids_directory / "sub-123" / "ses-456" / "ecephys" / "sub-123_ses-456_events.json"
    expected_json_content = {
        "onset": {"Description": "Onset of the event, measured from the beginning of the acquisition.", "Units": "s"},
        "duration": {"Description": "Duration of the event (measured from onset).", "Units": "s"},
        "trial_condition": {"Description": "Extra information per trial."},
        "nwb_table": {
            "Description": "The name of the NWB table from which this event was extracted.",
            "HED": {"trials": "Experimental-trial"},
            "Levels": {"trials": "The 'trials' table in the NWB file."},
        },
        "trials": {"Description": "A mock trials table."},
    }
    with json_file_path.open(mode="r") as file_stream:
        actual_json_content = json.load(fp=file_stream)
    assert actual_json_content == expected_json_content


def test_epochs_events(epochs_events_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    nwb_paths = [epochs_events_nwbfile_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
                "sub-123_ses-456_events.tsv",
                "sub-123_ses-456_events.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    json_file_path = temporary_bids_directory / "sub-123" / "ses-456" / "ecephys" / "sub-123_ses-456_events.json"
    expected_json_content = {
        "onset": {"Description": "Onset of the event, measured from the beginning of the acquisition.", "Units": "s"},
        "duration": {"Description": "Duration of the event (measured from onset).", "Units": "s"},
        "epoch_condition": {"Description": "Extra information per epoch."},
        "nwb_table": {
            "Description": "The name of the NWB table from which this event was extracted.",
            "HED": {"epochs": "Time-block"},
            "Levels": {"epochs": "The 'epochs' table in the NWB file."},
        },
        "epochs": {"Description": "A mock epochs table."},
    }
    with json_file_path.open(mode="r") as file_stream:
        actual_json_content = json.load(fp=file_stream)
    assert actual_json_content == expected_json_content


def test_multiple_events(multiple_events_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    nwb_paths = [multiple_events_nwbfile_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
                "sub-123_ses-456_events.tsv",
                "sub-123_ses-456_events.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    json_file_path = temporary_bids_directory / "sub-123" / "ses-456" / "ecephys" / "sub-123_ses-456_events.json"
    expected_json_content = {
        "onset": {"Description": "Onset of the event, measured from the beginning of the acquisition.", "Units": "s"},
        "duration": {"Description": "Duration of the event (measured from onset).", "Units": "s"},
        "trial_condition": {"Description": "Extra information per trial."},
        "epoch_condition": {"Description": "Extra information per epoch."},
        "tag": {"Description": "A tag assigned to each interval."},
        "nwb_table": {
            "Description": "The name of the NWB table from which this event was extracted.",
            "HED": {"epochs": "Time-block", "mock_time_intervals": "Time-interval", "trials": "Experimental-trial"},
            "Levels": {
                "epochs": "The 'epochs' table in the NWB file.",
                "mock_time_intervals": "The 'mock_time_intervals' table in the NWB file.",
                "trials": "The 'trials' table in the NWB file.",
            },
        },
        "trials": {"Description": "A mock trials table."},
        "epochs": {"Description": "A mock epochs table."},
        "mock_time_intervals": {"Description": "A mock time intervals table."},
    }
    with json_file_path.open(mode="r") as file_stream:
        actual_json_content = json.load(fp=file_stream)
    assert actual_json_content == expected_json_content



import pathlib
import typing

import pydantic
import pynwb
import pynwb.testing.mock.ecephys
import pynwb.testing.mock.file


def get_tutorial_directory() -> pathlib.Path:
    tutorial_dir = pathlib.Path.home() / "nwb2bids_tutorials"
    tutorial_dir.mkdir(exist_ok=True)
    return tutorial_dir


def _generate_ecephys_file(*, nwbfile_path: pathlib.Path, subject_id: str = "001", session_id: str = "A") -> None:
    nwbfile = pynwb.testing.mock.file.mock_NWBFile(
        session_id=session_id,
        session_description="An example NWB file containing ephys neurodata types - for use in the nwb2bids tutorials.",
    )

    subject = pynwb.file.Subject(
        subject_id=subject_id,
        species="Mus musculus",
        sex="M",
    )
    nwbfile.subject = subject

    pynwb.testing.mock.ecephys.mock_ElectricalSeries(
        name="ExampleElectricalSeries",
        description=(
            "An example electrical series that represents data which could have been "
            "read off of the channels of an ephys probe."
        ),
        nwbfile=nwbfile,
    )

    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)


@pydantic.validate_call
def generate_ephys_tutorial(
    *, mode=typing.Literal["file", "dataset"], output_directory: pydantic.DirectoryPath | None = None
) -> pathlib.Path:
    if output_directory is None:
        tutorial_dir = get_tutorial_directory()
        output_directory = tutorial_dir / f"ephys_tutorial_{mode}"
        output_directory.mkdir(exist_ok=True)

    if mode == "file":
        nwbfile_path = output_directory / "ephys.nwb"
        _generate_ecephys_file(nwbfile_path=nwbfile_path)

        return nwbfile_path
    elif mode == "dataset":
        subdir = output_directory / "some_sessions"
        subdir.mkdir(exist_ok=True)
        index_to_paths = {
            0: output_directory / subdir / "ephys_session_1.nwb",
            1: output_directory / subdir / "ephys_session_2.nwb",
            2: output_directory / "ephys_session_3.nwb",
            3: output_directory / "DO_NOT_CONVERT.nwb",
        }
        index_to_subject_id = {
            0: "001",
            1: "001",
            2: "002",
            3: "003",
        }
        index_to_session_id = {
            0: "A",
            1: "B",
            2: "C",
            3: "D",
        }

        for index in range(4):
            nwbfile_path = index_to_paths[index]
            _generate_ecephys_file(
                nwbfile_path=nwbfile_path, subject_id=index_to_subject_id[index], session_id=index_to_session_id[index]
            )

        return output_directory



import abc

import pydantic

from ._run_config import RunConfig


class BaseConverter(pydantic.BaseModel, abc.ABC):
    run_config: RunConfig = pydantic.Field(
        description="The configuration for this conversion run.", default_factory=RunConfig
    )

    @abc.abstractmethod
    def extract_metadata(self) -> None:
        """
        Extract essential metadata used by the BIDS standard from the source NWB files.
        """
        message = f"The `extract_metadata` method has not been implemented by the `{self.__class__.__name__}` class."
        raise NotImplementedError(message)



import json
import pathlib
from typing import Any

import pandas
import pydantic
import pynwb
import typing_extensions

from .._inspection._inspection_result import Category, DataStandard, InspectionResult, Severity
from ..bids_models._base_metadata_model import BaseMetadataContainerModel, BaseMetadataModel


class Probe(BaseMetadataModel):
    probe_id: str
    type: str | None = None
    description: str | None = None
    manufacturer: str | None = None


class ProbeTable(BaseMetadataContainerModel):
    probes: list[Probe]

    def _check_fields(self) -> None:
        # Check if values are specified
        self._internal_messages = []

        probes_missing_description = [probe for probe in self.probes if probe.description is None]
        for probe_missing_description in probes_missing_description:
            self._internal_messages.append(
                InspectionResult(
                    title="Missing description",
                    reason="A basic description of this field is recommended to improve contextual understanding.",
                    solution="Add a description to the field.",
                    field=f"nwbfile.devices.{probe_missing_description.probe_id}",
                    source_file_paths=[],  # TODO: figure out better way of handling
                    data_standards=[DataStandard.BIDS, DataStandard.NWB],
                    category=Category.STYLE_SUGGESTION,
                    severity=Severity.INFO,
                )
            )

    def model_post_init(self, context: Any, /) -> None:
        self._check_fields()

    @pydantic.computed_field
    @property
    def messages(self) -> list[InspectionResult]:
        """
        All messages from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        messages = [message for probe in self.probes for message in probe.messages]
        messages += self._internal_messages
        messages.sort(key=lambda message: (-message.category.value, -message.severity.value, message.title))
        return messages

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self | None:
        nwb_electrode_tables = [nwbfile.electrodes for nwbfile in nwbfiles]
        if not any(nwb_electrode_tables):
            return None

        unique_devices = {
            electrode_group.device
            for nwbfile in nwbfiles
            for electrode_group in nwbfile.electrodes["group"][:]
            if nwbfile.electrodes is not None
        }

        probes = [
            Probe(
                probe_id=device.name,
                type=None,  # TODO
                description=device.description,
                manufacturer=device.manufacturer,
            )
            for device in unique_devices
        ]
        return cls(probes=probes)

    @pydantic.validate_call
    def to_tsv(self, file_path: str | pathlib.Path) -> None:
        """
        Save the probes information to a TSV file.

        Parameters
        ----------
        file_path : path
            The path to the output TSV file.
        """
        data = []
        for probe in self.probes:
            model_dump = probe.model_dump()
            data.append(model_dump)

        data_frame = pandas.DataFrame(data=data)
        data_frame.to_csv(path_or_buf=file_path, sep="\t", index=False)

    @pydantic.validate_call
    def to_json(self, file_path: str | pathlib.Path) -> None:
        """
        Save the probes information to a JSON file.

        Parameters
        ----------
        file_path : path
            The path to the output JSON file.
        """
        with file_path.open(mode="w") as file_stream:
            json.dump(obj=dict(), fp=file_stream, indent=4)



import pathlib


def _read_first_bytes(file_path: pathlib.Path, n: int = 6) -> bytes:
    """
    Read the first n (by default 6) bytes of a file; necessary for Windows 'symlinks'.
    """
    with file_path.open(mode="rb") as file_stream:
        first_bytes = file_stream.read(n)
        return first_bytes


def _content_is_retrieved(file_path: pathlib.Path) -> bool:
    """
    Check if a file contents are fetched from the git-annex.

    Parameters
    ----------
    file_path : pathlib.Path
        Path to the file to check.

    Returns
    -------
    bool
        True if the file is a pointer to annex contents, False otherwise.
    """
    if not file_path.exists():
        return False

    # Note that if the file path is a 'broken' symlink (the target does not exist)
    # the call to `.exists()` would actually return False
    if file_path.is_symlink() and (".git" in file_path.readlink().parts):
        return True

    return _read_first_bytes(file_path=file_path, n=6) != b"/annex"



import pynwb
import pynwb.testing.mock.base


def mock_trials_table(nwbfile: pynwb.NWBFile | None = None) -> pynwb.epoch.TimeIntervals:
    """
    Generate a mock trials table.

    These are shorter than epochs and fit neatly within them in a qualitative manner.
    If a NWBFile is provided, a TimeSeriesReference to a mock TimeSeries will be added to each trial.
    """
    trials = pynwb.epoch.TimeIntervals(name="trials", description="A mock trials table.")
    trials.add_column(name="trial_condition", description="Extra information per trial.")
    trials.add_column(
        name="indexed_column",
        description="A more complicated type of column that contains structured data.",
        index=True,
    )

    kwargs_per_row = [
        {"start_time": 0.0, "stop_time": 1.0, "trial_condition": "A", "indexed_column": [0, 1]},
        {"start_time": 2.0, "stop_time": 3.0, "trial_condition": "B", "indexed_column": [2, 3]},
        {"start_time": 5.0, "stop_time": 5.5, "trial_condition": "C", "indexed_column": [4, 5]},
        {"start_time": 5.5, "stop_time": 6.0, "trial_condition": "D", "indexed_column": [6, 7]},
    ]
    if nwbfile is not None:
        timeseries = pynwb.testing.mock.base.mock_TimeSeries(nwbfile=nwbfile)
        timeseries_ref = pynwb.base.TimeSeriesReference(
            idx_start=0, count=timeseries.data.shape[0], timeseries=timeseries
        )

        for kwargs in kwargs_per_row:
            kwargs["timeseries"] = [timeseries_ref]
    for kwargs in kwargs_per_row:
        trials.add_row(**kwargs)

    return trials


def mock_epochs_table() -> pynwb.epoch.TimeIntervals:
    """
    Generate a mock epochs table.

    These are longer than trials and encompass several of them in a qualitative manner.
    """
    epochs = pynwb.epoch.TimeIntervals(name="epochs", description="A mock epochs table.")
    epochs.add_column(name="epoch_condition", description="Extra information per epoch.")
    epochs.add_row(start_time=0.0, stop_time=3.5, epoch_condition="AB")
    epochs.add_row(start_time=4.5, stop_time=6.0, epoch_condition="CD")

    return epochs


def mock_time_intervals() -> pynwb.epoch.TimeIntervals:
    """Generate a mock time intervals table."""
    time_intervals = pynwb.epoch.TimeIntervals(name="mock_time_intervals", description="A mock time intervals table.")
    time_intervals.add_column(name="tag", description="A tag assigned to each interval.")
    time_intervals.add_row(start_time=0.5, stop_time=0.75, tag="tag1")
    time_intervals.add_row(start_time=5.1, stop_time=5.23, tag="tag2")

    return time_intervals
